#include <stdarg.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <complex>
#include <string.h>

using namespace std;

/*
  This program is a C++ implementation of a Matlab routine by Michael
  Kleinschmidt to apply a specified set of Gabor filters to an input
  spectrogram.

  This program processes a single utterance each time it is invoked.

  Input and output must be in HTK feature file format. 
  
  THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND ANY EXPRESS 
  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
  DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*/

/* The following defines the parameters of a Gabor filter as
   understood by this program.  These parameters are read in from a
   parameter file whose name is specified as a command line argument
   to this program.  See Kleinschmidt's Ph.D. thesis or his other
   publications for more information. The term "F dimension" refers to
   the frequency axis of the input spectrogram and the term
   "T dimension" refers to the time axis of the input spectrogram. */

enum gabortype {REAL, IMAG, ABS};  /* Take real part, imaginary part,
              or absolute magnitude of output. */
struct gabor
{
  enum gabortype type;
  int f0;        /*  Center frequency assigned to the Gabor filter. This program
       will center the Gabor filter on this center frequency when
       applying it to the spectrogram. */
  double FmodF;  /* Ripple frequency (F dimension) */
  double TmodF;  /* Ripple velocity (T dimension) */
  double Fsigma; /* Variance of envelope in F dimension relative
       to period of Fmod */
  double Tsigma; /* Variance of envelope in T dimension relative
        to period of Tmod */
  int Fcut;      /* Cut-off the envelope at this many points from the center  (F dimension) */
  int Tcut;      /* Cut-off the envelope at this many points from the center  (T dimension)*/
  complex<double> *func; /* This is used to store the calculated Gabor filter. */

};

/* This toggles the display of status messages. */
bool bVerboseFlag = 1;

/* Header of an HTK feature file. */
struct htk_header {
  int nSamples;   /* "number of samples in file" */
  int sampPeriod; /* "sample period in 100ns units" */
  short sampSize; /* "number of bytes per sample" */
  short sampKind; /* "a code indicating the sample kind" */
} header;

/* This function prints an error message and exits. */
void
fatal(char *s, ...) {
  va_list ap;
  va_start(ap, s);
  vfprintf(stderr, s, ap);
  fputc('\n', stderr);
  exit(1);
}

/* This function prints a status message. */
void
status(char *s, ...) {
  va_list ap;
  va_start(ap, s);
  vprintf(s, ap);
  fflush(stdout);
}

/* This function calculates Gabor filters from the parameters that
   define them.  For more information, see Kleinschmidt's thesis or
   his papers, e.g., "Methods for capturing spectro-temporal
   modulations in automatic speech recognition", ACUSTICA * acta
   acustica August 2001.   The first argument to this function
   (gp) points to an array of type struct gabor, and the
   second argument (ng) gives the length of that array. */
void create_gabor(struct gabor *gp, int ng, char *envtype)
{
  int i;
  bool bUseHannEnv = 0;
  bool bUseGausEnv = 0; 

  // check for valid type of envelope
  if (strcmp(envtype,"HANNING") == 0)
    bUseHannEnv = 1;
  if (strcmp(envtype,"GAUSSIAN") == 0)
    bUseGausEnv = 1;
  if ((!bUseHannEnv) && (!bUseGausEnv))
    fatal("Type of envelope should be GAUSSIAN or HANNING\n");  

  for (i=0; i<ng; i++)
  {
    struct gabor *g;
    int nRows, nCols, row, col;
    int f,t ;
    int f_correct_count, t_correct_count;
    const double ValPi = 3.141592653589;
    double sum_hanning, scale_hanning;
    double envelope_hanning;

    g = &gp[i];

    nRows = 2*g->Fcut+1;
    nCols = 2*g->Tcut+1;      

    g->func = new complex<double>[nRows * nCols];
    if (!g->func)
      fatal("Unable to allocate memory for Gabor function.");

    if (bUseHannEnv)
    {
      // -----------------------------------------------
      // calculate hanning scale factor
      // -----------------------------------------------
      
      envelope_hanning = 0;
      sum_hanning      = 0;
      scale_hanning    = 0;

      row = 0;
      f_correct_count = 0;

      for (f = -g->Fcut; f <= g->Fcut; f += 1)
      {
        f_correct_count = f_correct_count + 1;
        col = 0;
        t_correct_count = 0;
        for (t = -g->Tcut; t <= g->Tcut; t += 1)
        {
          double hanning_w, hanning_t;
          hanning_w = 0;
          hanning_t = 0;
          t_correct_count = t_correct_count + 1;
          /* Hanning envelope */
          hanning_w = 0.5 - 0.5*cos(2*ValPi*f_correct_count / (nRows+1));
          hanning_t = 0.5 - 0.5*cos(2*ValPi*t_correct_count / (nCols+1));
          envelope_hanning = hanning_w * hanning_t;
          sum_hanning = envelope_hanning + sum_hanning;        
          col++;
        }  
        row++;
      }
      scale_hanning = 1/sum_hanning;
      envelope_hanning = 0;

    }

    row = 0;
    f_correct_count = 0;

    for (f = -g->Fcut; f <= g->Fcut; f += 1) 
    {
      f_correct_count = f_correct_count + 1;
      t_correct_count = 0;    
      col = 0;    
      for (t = -g->Tcut; t <= g->Tcut; t += 1) 
      {

        double scale, gauss_w, gauss_t, envelope;
        double hanning_w = 0.0;
        double hanning_t = 0.0;
        complex<double> euler;
        complex<double> j(0,1);

        if (bUseGausEnv)
        {
          /* Gaussian envelope */
          if  ((bVerboseFlag == 1) && (i == 0) && (f == -g->Fcut) && (t == -g->Tcut))
            status("Using Gaussian envelope\n");

          scale = 1/(2*ValPi*g->Fsigma*g->Tsigma);
          gauss_w = (-(f*f)) / (2*(g->Fsigma*g->Fsigma));
          gauss_t = (-(t*t)) / (2*(g->Tsigma*g->Tsigma));
          envelope = scale*exp(gauss_w+gauss_t);
        }
        
        if (bUseHannEnv)
        {
          /* Hanning envelope */
          t_correct_count = t_correct_count + 1;

          if  ((bVerboseFlag == 1) && (i == 0) && (f == -g->Fcut) && (t == -g->Tcut))
            status("Using Hanning envelope\n");

          hanning_w = 0.5 - 0.5*cos(2*ValPi*f_correct_count / (nRows+1));
          hanning_t = 0.5 - 0.5*cos(2*ValPi*t_correct_count / (nCols+1));
          envelope = scale_hanning * hanning_w * hanning_t;
        }
        
        /* Complex Euler function */
        euler = exp(j*((double) 2*ValPi)*(g->FmodF*f + g->TmodF*t));

        g->func[row*nCols + col] = envelope * euler;
    
        col++;
      }
    row++;
  }

  // REAL and IMAG Gabor functions need to have mean removed
  // to eliminate DC bias.  Furthermore since the input
  // feature matrix is real we can simply replace the
  // Gabor function with its real or imaginary part
  // to save computation.
      
  if (g->type == REAL || g->type == IMAG)
  {
    double mean;

    for (f = 0; f < nRows*nCols; f++)
    {
      if (g->type == REAL)
      // Set the imaginary part to zero.
      g->func[f] = complex<double>(g->func[f].real(), 0);
      else // IMAG
      // Set the real part to the imaginary part
      // and set the imaginary part to zero (this
      // this way we don't need to keep track 
      // later on of whether we are in the REAL or IMAG case).
      g->func[f] = complex<double>(g->func[f].imag(), 0);
    }
    
    // Take away the mean
    mean = 0.0;
    for (f = 0; f < nRows*nCols; f++)
      mean += g->func[f].real();
    mean = mean / (nRows*nCols);
    for (f = 0; f < nRows*nCols; f++)
      g->func[f] -= mean;
    }
  }
}

/* Loads Gabor filter parameter definitions from file. */
int read_gabor(char *filename, struct gabor **pp)
{
  FILE *fp;
  int ct = 0;
  int ct2 = 0;
  char buf[1024];
  struct gabor *p;

  /* Open file for read. */
  if ((fp = fopen(filename,"r")) == NULL)
    fatal("Cannot open file %s for reading.", filename);

  /* Count number of non-comment lines. */
  while (fgets(buf, sizeof(buf), fp) != NULL) {
    if (buf[0] != '#') {
      ct++;
    }
  }

  /* Move back to start of file. */
  rewind(fp); 

  /* Allocate array of Gabor function parameter structs. */
  p = (struct gabor *) malloc( sizeof(struct gabor) * ct );
  if (!p) {
    fatal("Unable to allocate Gabor parameter array.");
  }
  
  /* Now read in the parameters. */
  ct2 = 0; 
  while (fgets(buf, sizeof(buf), fp) != NULL) {
    if (buf[0] != '#') {
      int nvals;
      int countfield;
      char type[1024];

      nvals = sscanf(buf, "%d %d %lf %lf %lf %lf %d %d %s",
         &countfield, 
         &p[ct2].f0, &p[ct2].FmodF, &p[ct2].TmodF, 
         &p[ct2].Fsigma, &p[ct2].Tsigma,
         &p[ct2].Fcut, &p[ct2].Tcut, type);

      if (!strcmp(type, "real"))
  p[ct2].type = REAL;
      else if (!strcmp(type, "imag"))
  p[ct2].type = IMAG;
      else if (!strcmp(type, "abs") || !strcmp(type, "mag"))
  p[ct2].type = ABS;
      else 
  fatal("Unknown type %s in Gabor function definition %s.", type, buf);


      if (nvals != 9) 
  fatal("Parse error reading Gabor function definition: %s.", buf);

      ct2++;
    }
  }

  if (ct != ct2)
    fatal("I/O error reading Gabor function definitions");

  *pp = p;  
  return ct;
}

void write_htk_file( char * filename, float *data) 
{
  FILE *fp;

  /* Open file for write. */
  if ((fp = fopen(filename,"wb")) == NULL)
    fatal("Cannot open file %s for writing.", filename);

  if(fwrite(&header, sizeof(struct htk_header), 1, fp) != 1) {
    fatal("Cannot write HTK header to file %s.", filename);
  }

  if((int) fwrite(data, header.sampSize, header.nSamples, fp) != header.nSamples) {
    fatal("Cannot write all data to file %s.", filename);
  }
    
  fclose(fp);
}


void read_htk_file( char * filename, float **data) 
{
  FILE *fp;
  unsigned long int length;

  /* Open file for read. */
  if ((fp = fopen(filename,"rb")) == NULL)
    fatal("Cannot open file %s for reading.", filename);

  if(fread(&header, sizeof(struct htk_header), 1, fp) != 1) {
    fclose(fp);
    fatal("Cannot read HTK header from file %s.", filename);
  }
  
  if (bVerboseFlag)
  {
    status("Input HTK header:\n");
    status("  Number of frames (called 'samples' in HTK manual): %d\n",
	   header.nSamples);
    status("  Frame period: %d (%f milliseconds)\n",
	   header.sampPeriod, ((float) header.sampPeriod/10000));
    status("  Frame size (bytes):  %d\n", header.sampSize);
    status("  Frame size (floats): %d\n", header.sampSize/4);
    status("  Frame kind:   %d\n",  header.sampKind);
  }
  
  length = header.nSamples*header.sampSize/4;

  *data = (float *) malloc(sizeof(float) * length);

  if(!*data) {
    fatal("Memory allocation error in read_htk_file");
  }

  if(fread(*data, sizeof(float), length, fp) != length) {
    fatal("Read less than %lu floats from %s", length, filename);
  }

  if (bVerboseFlag)
    {
      int i;

      status("First frame of input data: ");
      for (i = 0; i < header.sampSize/4; i++) {
		status("%f ", (*data)[i]);
      }
      status("\n");
    }
  
  fclose(fp);
}

/* Reads in input feature file.  Adds padding to edges of feature
   matrix. */
void read_htk_file_and_pad( char * filename, int Fcut, int Tcut,
          float **data,
         
          int *nRows, int *nCols)
{
  float *filedata;
  float *p;
  int oldrow, oldcol, row, col;
  int nRowsBeforePad, nColsBeforePad;

  read_htk_file(filename, &filedata);
 
  nRowsBeforePad = header.nSamples;
  nColsBeforePad = header.sampSize/4;

  *nRows = nColsBeforePad + 2*Fcut; /* rows will be frequency dimn */
  *nCols = nRowsBeforePad + 2*Tcut; /* cols will be time dimn */

  /* Allocate storage for features with padding. */
  p = (float *) malloc( (*nRows) * (*nCols) * sizeof(float));
  if(!p) {
    fatal("Memory allocation error in read_htk_file");
  }
  *data = p;  

  /* Copy features from filedata into p. */
  for (oldrow = 0; oldrow < nRowsBeforePad; oldrow++)
    for (oldcol= 0; oldcol < nColsBeforePad; oldcol++)
      {

  /* At this point in filedata[] we have 
     
     rows = time dimension 
     cols = frequency dimension
     
     When we do the copy into p, we switch rows and columns in order to 
     tranpose the matrix, and we end up with

     rows = frequency dimn
     cols = time dimn 

           */
  
  row = oldcol + Fcut;
  col = oldrow + Tcut;


  p[row*(*nCols) + col] = filedata[oldrow*(nColsBeforePad) + oldcol];
      }
  
  /* Put padding into p.  We duplicate the edge features from the file
     on each side to create padding.  Unlike padding with zeros, this
     does not cause onset artifacts. */

  /* Left side pad. */
  for (row = 0; row < *nRows; row++)
    for (col = 0; col < Tcut; col++) 
      p[row*(*nCols) + col] = p[row*(*nCols) + Tcut];

  /* Right side pad. */
  for (row = 0; row < *nRows; row++)
    for (col = *nCols - Tcut; col < *nCols; col++) 
      p[row*(*nCols) + col] = p[row*(*nCols) + *nCols -Tcut - 1];

  /* Top side pad. */
  for (row = 0; row < Fcut; row++)
    for (col = 0; col < *nCols; col++)
      p[row*(*nCols) + col] = p[Fcut*(*nCols) + col];

  /* Bottom side pad. */
  for (row = *nRows - Fcut; row < *nRows; row++)
    for (col = 0; col < *nCols; col++)
      p[row*(*nCols) + col] = p[(*nRows -Fcut - 1)*(*nCols) + col];

  /* Free the storage for the original (unpadded) matrix. */
  free(filedata);
}

int main(int argc, char *argv[])
{
  /* Gabor filter parameter file, input feature file,
    output feature file. */  
  char *gaborparam, *envtype, *infile, *outfile;

  /* Input feature matrix. */
  float *data;

  /* Output feature matrix. */
  float *outdata;

  /* Pointer to array of Gabor filter definitions. */
  struct gabor *gp;

  /* Number of Gabor filter definitions. */
  int ng;

  /* Number of rows and columns for input feature matrix, including padding. */  
  int nRows, nCols; 

  int Tpad, Fpad; 
  int i;
  int frame, nFrames;

  if ((argc < 5) || (argc > 7))
    fatal("Usage: gabor [-q] <gaborparam> <type of envelope> <infile> <outfile>");

  
  /* Check for flag for quiet (non-verbose) mode. */
  if (strcmp(argv[1], "-q") == 0) {
    bVerboseFlag = 0;
    gaborparam = argv[2];
    envtype    = argv[3];
    infile     = argv[4];
    outfile    = argv[5];
  }
  else {
    gaborparam = argv[1];
    envtype    = argv[2];
    infile     = argv[3];
    outfile    = argv[4];
  }

  /* Read in Gabor filter parameter definitions. */
  ng = read_gabor(gaborparam, &gp);

  if (ng==0)
    fatal("\nNo Gabor functions defined in %s; exiting.", gaborparam);

  /* Create the Gabor filters. */
  create_gabor(gp, ng, envtype);

  /* Find max Fcut, Tcut.  Use that to determine how much we need
     to pad the input feature matrix. */
  Fpad=gp[0].Fcut;
  Tpad=gp[0].Tcut;
  for (i = 1; i < ng; i++){
    if (gp[i].Fcut > Fpad)
      Fpad = gp[i].Fcut;
    if (gp[i].Tcut > Tpad)
      Tpad = gp[i].Tcut;
  }
  
  /* Read and pad the input feature matrix. */
  read_htk_file_and_pad(infile, Fpad, Tpad, &data, &nRows, &nCols);

  /* One frame of output for each frame of input not including
     padding */
  nFrames = nCols -  2*Tpad;

  /* Allocate output storage. */
  outdata = (float *) malloc( nFrames * ng * (sizeof(float)));
  if (!outdata)
    fatal("Unable to allocate output array.");

  /* For each Gabor function. */
  for (i = 0; i < ng; i++)
    {
      
      /* For each frame of output. */
      for (frame = 0; frame < nFrames; frame++)
	{
	  /* Row and column iteration variables for doing the 2-D
	     sum-of-products between the input feature matrix and the
	     Gabor filter.  r and c are row and column for input
	     feature matrix.  fr and fc are row and column for Gabor
	     filter matrix.  The number of columns in the Gabor filter matrix is
	     2*gp[i].Tcut + 1. */
	  int r, c, fr, fc; 
	  
	  complex<double> sum, rowsum;
	  float outval;
	  
	  /* The Gabor filter is applied to the input matrix centered
	     on the current input frame and the Gabor filter's center
	     frequency f0. In the Gabor filter definition, f0 is
	     specified by counting bins starting from 1 (matlab
	     style), so we subtract 1 from it in order to count from
	     0. */
    
	  sum = 0;
	  fr = 0;
	  for (r = Fpad + gp[i].f0 - 1 - gp[i].Fcut; 
	       r <= Fpad + gp[i].f0 - 1 + gp[i].Fcut; r++)
	    { 
	      rowsum = 0;

	      fc = 0;
	      for (c = Tpad + frame - gp[i].Tcut; 
		   c <= Tpad + frame + gp[i].Tcut; c++)
		{
		  rowsum += 
		    ((double) data[r*nCols + c]) * gp[i].func[fr*(2*gp[i].Tcut+1) + fc];
		  fc++;
		}
	      sum = sum + rowsum;
	      fr++;
	    }
    
	  if (gp[i].type == ABS) 
	    outval = sqrt(sum.real()*sum.real()+sum.imag()*sum.imag());
	  else // REAL or IMAG
	    outval = sum.real();

	  outdata[frame*ng + i] = outval;
	}
    }
  
  header.nSamples = nFrames;
  header.sampSize = ng * 4;
  header.sampKind = 9; /* This is the code for a user-defined
			  feature format, according to 'HTK Format
			  Parameter Files' in the HTK manual. */

  if (bVerboseFlag)
    {
      status("Output HTK header:\n");
      status("  Number of frames (called 'samples' in HTK manual): %d\n",
	     header.nSamples);
      status("  Frame period: %d (%f milliseconds)\n",
	     header.sampPeriod, ((float) header.sampPeriod/10000));
      status("  Frame size (bytes):  %d\n", header.sampSize);
      status("  Frame size (floats): %d\n", header.sampSize/4);
      status("  Frame kind:   %d\n",  header.sampKind);
      
      status("First frame of output data: ");
      for (i = 0; i < ng; i++) {
	status("%f ", outdata[i]);
      }
      status("\n");
    }
  
  write_htk_file(outfile, outdata);
}
